/*
	SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2004 Sam Lantinga

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.

	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Sam Lantinga
	slouken@libsdl.org
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_fbevents.c,v 1.13 2004/01/04 16:49:25 slouken Exp $";
#endif

/* Handle the event stream, converting console events into SDL events */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

/* For parsing /proc */
#include <dirent.h>
#include <ctype.h>

#include <linux/vt.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>

#include <linux/input.h>

/* D. Finkler */
#include "KeyCodes.h"

#include "SDL.h"
#include "SDL_mutex.h"
#include "SDL_sysevents.h"
#include "SDL_sysvideo.h"
#include "SDL_events_c.h"
#include "SDL_fbvideo.h"
#include "SDL_fbevents_c.h"
#include "SDL_fbkeys.h"

#include "SDL_fbelo.h"

#ifndef GPM_NODE_FIFO
#define GPM_NODE_FIFO	"/dev/gpmdata"
#endif


/* The translation tables from a console scancode to a SDL keysym */
#define NUM_VGAKEYMAPS	(1<<KG_CAPSSHIFT)
static Uint16 vga_keymap[NUM_VGAKEYMAPS][NR_KEYS];
static SDLKey keymap[128];
static Uint16 keymap_temp[128]; /* only used at startup */
//static SDL_keysym *TranslateKey(int scancode, SDL_keysym *keysym);

/* Ugh, we have to duplicate the kernel's keysym mapping code...
   Oh, it's not so bad. :-)

   FIXME: Add keyboard LED handling code
 */
static void FB_vgainitkeymaps(int fd)
{
	struct kbentry entry;
	int map, i;

	/* Don't do anything if we are passed a closed keyboard */
	if ( fd < 0 ) {
		return;
	}

	/* Load all the keysym mappings */
	for ( map=0; map<NUM_VGAKEYMAPS; ++map ) {
		memset(vga_keymap[map], 0, NR_KEYS*sizeof(Uint16));
		for ( i=0; i<NR_KEYS; ++i ) {
			entry.kb_table = map;
			entry.kb_index = i;
			if ( ioctl(fd, KDGKBENT, &entry) == 0 ) {
				/* fill keytemp. This replaces SDL_fbkeys.h */
				if ( (map == 0) && (i<128) ) {
					keymap_temp[i] = entry.kb_value;
				}
				/* The "Enter" key is a special case */
				if ( entry.kb_value == K_ENTER ) {
					entry.kb_value = K(KT_ASCII,13);
				}
				/* Handle numpad specially as well */
				if ( KTYP(entry.kb_value) == KT_PAD ) {
					switch ( entry.kb_value ) {
					case K_P0:
					case K_P1:
					case K_P2:
					case K_P3:
					case K_P4:
					case K_P5:
					case K_P6:
					case K_P7:
					case K_P8:
					case K_P9:
						vga_keymap[map][i]=entry.kb_value;
						vga_keymap[map][i]+= '0';
						break;
										case K_PPLUS:
						vga_keymap[map][i]=K(KT_ASCII,'+');
						break;
										case K_PMINUS:
						vga_keymap[map][i]=K(KT_ASCII,'-');
						break;
										case K_PSTAR:
						vga_keymap[map][i]=K(KT_ASCII,'*');
						break;
										case K_PSLASH:
						vga_keymap[map][i]=K(KT_ASCII,'/');
						break;
										case K_PENTER:
						vga_keymap[map][i]=K(KT_ASCII,'\r');
						break;
										case K_PCOMMA:
						vga_keymap[map][i]=K(KT_ASCII,',');
						break;
										case K_PDOT:
						vga_keymap[map][i]=K(KT_ASCII,'.');
						break;
					default:
						break;
					}
				}
				/* Do the normal key translation */
				if ( (KTYP(entry.kb_value) == KT_LATIN) ||
					 (KTYP(entry.kb_value) == KT_ASCII) ||
					 (KTYP(entry.kb_value) == KT_LETTER) ) {
					vga_keymap[map][i] = entry.kb_value;
				}
			}
		}
	}
}

int FB_InGraphicsMode(_THIS)
{
	return((keyboard_fd >= 0) && (saved_kbd_mode >= 0));
}

int FB_EnterGraphicsMode(_THIS)
{
	struct termios keyboard_termios;

	/* Set medium-raw keyboard mode */
	if ( (keyboard_fd >= 0) && !FB_InGraphicsMode(this) ) {

		/* Switch to the correct virtual terminal */
		if ( current_vt > 0 ) {
			struct vt_stat vtstate;

			if ( ioctl(keyboard_fd, VT_GETSTATE, &vtstate) == 0 ) {
				saved_vt = vtstate.v_active;
			}
			if ( ioctl(keyboard_fd, VT_ACTIVATE, current_vt) == 0 ) {
				ioctl(keyboard_fd, VT_WAITACTIVE, current_vt);
			}
		}

		/* Set the terminal input mode */
		if ( tcgetattr(keyboard_fd, &saved_kbd_termios) < 0 ) {
			SDL_SetError("Unable to get terminal attributes");
			if ( keyboard_fd > 0 ) {
				close(keyboard_fd);
			}
			keyboard_fd = -1;
			return(-1);
		}
		if ( ioctl(keyboard_fd, KDGKBMODE, &saved_kbd_mode) < 0 ) {
			SDL_SetError("Unable to get current keyboard mode");
			if ( keyboard_fd > 0 ) {
				close(keyboard_fd);
			}
			keyboard_fd = -1;
			return(-1);
		}
		keyboard_termios = saved_kbd_termios;
		keyboard_termios.c_lflag &= ~(ICANON | ECHO | ISIG);
		keyboard_termios.c_iflag &= ~(ISTRIP | IGNCR | ICRNL | INLCR | IXOFF | IXON);
		keyboard_termios.c_cc[VMIN] = 0;
		keyboard_termios.c_cc[VTIME] = 0;
		if (tcsetattr(keyboard_fd, TCSAFLUSH, &keyboard_termios) < 0) {
			FB_CloseKeyboard(this);
			SDL_SetError("Unable to set terminal attributes");
			return(-1);
		}
		/* This will fail if we aren't root or this isn't our tty */
		if ( ioctl(keyboard_fd, KDSKBMODE, K_MEDIUMRAW) < 0 ) {
			FB_CloseKeyboard(this);
			SDL_SetError("Unable to set keyboard in raw mode");
			return(-1);
		}
		if ( ioctl(keyboard_fd, KDSETMODE, KD_GRAPHICS) < 0 ) {
			FB_CloseKeyboard(this);
			SDL_SetError("Unable to set keyboard in graphics mode");
			return(-1);
		}
	}
	return(keyboard_fd);
}

void FB_LeaveGraphicsMode(_THIS)
{
	if ( FB_InGraphicsMode(this) ) {
		ioctl(keyboard_fd, KDSETMODE, KD_TEXT);
		ioctl(keyboard_fd, KDSKBMODE, saved_kbd_mode);
		tcsetattr(keyboard_fd, TCSAFLUSH, &saved_kbd_termios);
		saved_kbd_mode = -1;

		/* Head back over to the original virtual terminal */
		if ( saved_vt > 0 ) {
			ioctl(keyboard_fd, VT_ACTIVATE, saved_vt);
		}
	}
}

void FB_CloseKeyboard(_THIS)
{
	int usbkb;

	for (usbkb = 0; usbkb < 2; ++usbkb) {
		if (usbkeyboard_fd[usbkb] >= 0) {

			struct input_event evl;
			evl.type = EV_LED;
			evl.code = LED_SCROLLL;
			evl.value = 1;
			write(usbkeyboard_fd[usbkb], &evl, sizeof(evl));	

			close (usbkeyboard_fd[usbkb]);
			usbkeyboard_fd[usbkb] = -1;
		}
	}

	if ( keyboard_fd >= 0 ) {
		FB_LeaveGraphicsMode(this);
		if ( keyboard_fd > 0 ) {
			close(keyboard_fd);
		}
	}
	keyboard_fd = -1;
}

#if 0
int FB_OpenKeyboard(_THIS)
{
	/* Open only if not already opened */
 	if ( keyboard_fd < 0 ) {
		static const char * const tty0[] = { "/dev/tty0", "/dev/vc/0", NULL };
		static const char * const vcs[] = { "/dev/vc/%d", "/dev/tty%d", NULL };
		int i, tty0_fd;

		/* Try to query for a free virtual terminal */
		tty0_fd = -1;
		for ( i=0; tty0[i] && (tty0_fd < 0); ++i ) {
			tty0_fd = open(tty0[i], O_WRONLY, 0);
		}
		if ( tty0_fd < 0 ) {
			tty0_fd = dup(0); /* Maybe stdin is a VT? */
		}
		ioctl(tty0_fd, VT_OPENQRY, &current_vt);
		close(tty0_fd);
		if ( (geteuid() == 0) && (current_vt > 0) ) {
			for ( i=0; vcs[i] && (keyboard_fd < 0); ++i ) {
				char vtpath[12];

				sprintf(vtpath, vcs[i], current_vt);
				keyboard_fd = open(vtpath, O_RDWR, 0);
#ifdef DEBUG_KEYBOARD
				fprintf(stderr, "vtpath = %s, fd = %d\n",
					vtpath, keyboard_fd);
#endif /* DEBUG_KEYBOARD */

				/* This needs to be our controlling tty
				   so that the kernel ioctl() calls work
				*/
				if ( keyboard_fd >= 0 ) {
					tty0_fd = open("/dev/tty", O_RDWR, 0);
					if ( tty0_fd >= 0 ) {
						ioctl(tty0_fd, TIOCNOTTY, 0);
						close(tty0_fd);
					}
				}
			}
		}
 		if ( keyboard_fd < 0 ) {
			/* Last resort, maybe our tty is a usable VT */
			current_vt = 0;
			keyboard_fd = open("/dev/tty", O_RDWR);
 		}
#ifdef DEBUG_KEYBOARD
		fprintf(stderr, "Current VT: %d\n", current_vt);
#endif
 		saved_kbd_mode = -1;

		/* Make sure that our input is a console terminal */
		{ int dummy;
		  if ( ioctl(keyboard_fd, KDGKBMODE, &dummy) < 0 ) {
			close(keyboard_fd);
			keyboard_fd = -1;
			SDL_SetError("Unable to open a console terminal");
		  }
		}

		/* Set up keymap */
		FB_vgainitkeymaps(keyboard_fd);
 	}
 	return(keyboard_fd);
}
#endif

int FB_OpenKeyboard(_THIS)
{
	//fprintf(stderr, "FB_OpenKeyboard\n");

	char			devname[256] = "/dev/input/event";
	int				devnum;
	char *			devnump = devname + strlen(devname);
	int				usbkb = 0;
	unsigned short	id[4];
	int				fd;
	__s32			device = 1;	// Defaults to Hifidelio basic

	// USB keyboard
	while (usbkeyboard_fd[usbkb] >= 0 && usbkb < 2)
		usbkb++;

	for (devnum = 0; devnum < 32 && usbkb < 2; devnum++) {
		sprintf(devnump, "%d", devnum);
		
		if ((fd = open(devname, O_RDWR|O_NONBLOCK)) >= 0) {
			#ifdef DEBUG_KEYBOARD
				fprintf(stderr, "Open event device %s\n", devname);
			#endif

			if (ioctl(fd, EVIOCGID, id) >= 0 && id[ID_VENDOR] == 0x099a && id[ID_PRODUCT] == 0x610c) {

				struct input_event evl;
				evl.type = EV_LED;
				evl.code = LED_SCROLLL;
				evl.value = 1;
				write(fd, &evl, sizeof(evl));	

				usbkeyboard_fd[usbkb++] = fd;
				#ifdef DEBUG_KEYBOARD
					fprintf(stderr, "Found USB Keyboard %d on %s\n", usbkb, devname);
				#endif
				while (usbkeyboard_fd[usbkb] >= 0 && usbkb < 2)
					usbkb++;
				if (usbkb >= 2) break;
			}
			else {
				#ifdef DEBUG_KEYBOARD
					fprintf(stderr, "Close input device %s\n", devname);
				#endif
				close(fd);
			}
		}
	}

	/* Open only if not already opened */
	if ( keyboard_fd < 0 ) {
		static const char * const i2c[] = { "/dev/i2c-0", "/dev/i2c", NULL };
		int i, i2c0_fd;

		/* Try to query for a free i2c device */
		i2c0_fd = -1;
		for ( i=0; i2c[i] && (i2c0_fd < 0); ++i ) {
			i2c0_fd = open(i2c[i], O_RDWR, 0);
		}

		if (i2c0_fd < 0)
			SDL_SetError("Unable to open a i2c input device");

		/* D. Finkler: get device ID to get address of front panel controller */
		if (ioctl(i2c0_fd, I2C_SLAVE, 0x53) < 0)
			SDL_SetError("Unable to set slave address 0x53");
			
		for (i = 0; i < 10; ++i)
		{
			if ((device = i2c_smbus_read_byte_data (i2c0_fd, 0xfe)) >= 0)
				break;
			usleep (10000);
		}

		#ifdef DEBUG_KEYBOARD
			fprintf(stderr, "Device ID: %d\n", device);
		#endif
		
		/* Sputnik */
		if (device == 3)
		{
			if (ioctl(i2c0_fd, I2C_SLAVE, 0x14) < 0)
				SDL_SetError("Unable to set slave address 0x14");
		}
		/* Hifidelio */
		else		
		{
			if (ioctl(i2c0_fd, I2C_SLAVE, 0x10) < 0)
				SDL_SetError("Unable to set slave address 0x10");
		}
		
		keyboard_fd = i2c0_fd;

		/* Set up keymap */
		FB_vgainitkeymaps(keyboard_fd);
	}
	return(keyboard_fd);
}

static enum {
	MOUSE_NONE = -1,
	MOUSE_MSC,	/* Note: GPM uses the MSC protocol */
	MOUSE_PS2,
	MOUSE_IMPS2,
	MOUSE_MS,
	MOUSE_BM,
	MOUSE_ELO,
	NUM_MOUSE_DRVS
} mouse_drv = MOUSE_NONE;

void FB_CloseMouse(_THIS)
{
	if ( mouse_fd > 0 ) {
		close(mouse_fd);
	}
	mouse_fd = -1;
}

/* Returns processes listed in /proc with the desired name */
static int find_pid(DIR *proc, const char *wanted_name)
{
	struct dirent *entry;
	int pid;

	/* First scan proc for the gpm process */
	pid = 0;
	while ( (pid == 0) && ((entry=readdir(proc)) != NULL) ) {
		if ( isdigit(entry->d_name[0]) ) {
			FILE *status;
			char path[PATH_MAX];
			char name[PATH_MAX];

			sprintf(path, "/proc/%s/status", entry->d_name);
			status=fopen(path, "r");
			if ( status ) {
				name[0] = '\0';
				fscanf(status, "Name: %s", name);
				if ( strcmp(name, wanted_name) == 0 ) {
					pid = atoi(entry->d_name);
				}
				fclose(status);
			}
		}
	}
	return pid;
}

/* Returns true if /dev/gpmdata is being written to by gpm */
static int gpm_available(void)
{
	int available;
	DIR *proc;
	int pid;
	int cmdline, len, arglen;
	char path[PATH_MAX];
	char args[PATH_MAX], *arg;

	/* Don't bother looking if the fifo isn't there */
	if ( access(GPM_NODE_FIFO, F_OK) < 0 ) {
		return(0);
	}

	available = 0;
	proc = opendir("/proc");
	if ( proc ) {
		while ( (pid=find_pid(proc, "gpm")) > 0 ) {
			sprintf(path, "/proc/%d/cmdline", pid);
			cmdline = open(path, O_RDONLY, 0);
			if ( cmdline >= 0 ) {
				len = read(cmdline, args, sizeof(args));
				arg = args;
				while ( len > 0 ) {
					if ( strcmp(arg, "-R") == 0 ) {
						available = 1;
					}
					arglen = strlen(arg)+1;
					len -= arglen;
					arg += arglen;
				}
				close(cmdline);
			}
		}
		closedir(proc);
	}
	return available;
}


/* rcg06112001 Set up IMPS/2 mode, if possible. This gives
 *  us access to the mousewheel, etc. Returns zero if
 *  writes to device failed, but you still need to query the
 *  device to see which mode it's actually in.
 */
static int set_imps2_mode(int fd)
{
	/* If you wanted to control the mouse mode (and we do :)  ) ...
		Set IMPS/2 protocol:
			{0xf3,200,0xf3,100,0xf3,80}
		Reset mouse device:
			{0xFF}
	*/
	Uint8 set_imps2[] = {0xf3, 200, 0xf3, 100, 0xf3, 80};
	Uint8 reset = 0xff;
	fd_set fdset;
	struct timeval tv;
	int retval = 0;

	if ( write(fd, &set_imps2, sizeof(set_imps2)) == sizeof(set_imps2) ) {
		if (write(fd, &reset, sizeof (reset)) == sizeof (reset) ) {
			retval = 1;
		}
	}

	/* Get rid of any chatter from the above */
	FD_ZERO(&fdset);
	FD_SET(fd, &fdset);
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	while ( select(fd+1, &fdset, 0, 0, &tv) > 0 ) {
		char temp[32];
		read(fd, temp, sizeof(temp));
	}

	return retval;
}


/* Returns true if the mouse uses the IMPS/2 protocol */
static int detect_imps2(int fd)
{
	int imps2;

	imps2 = 0;

	if ( getenv("SDL_MOUSEDEV_IMPS2") ) {
		imps2 = 1;
	}
	if ( ! imps2 ) {
		Uint8 query_ps2 = 0xF2;
		fd_set fdset;
		struct timeval tv;

		/* Get rid of any mouse motion noise */
		FD_ZERO(&fdset);
		FD_SET(fd, &fdset);
		tv.tv_sec = 0;
		tv.tv_usec = 0;
		while ( select(fd+1, &fdset, 0, 0, &tv) > 0 ) {
			char temp[32];
			read(fd, temp, sizeof(temp));
		}

   		/* Query for the type of mouse protocol */
   		if ( write(fd, &query_ps2, sizeof (query_ps2)) == sizeof (query_ps2)) {
   			Uint8 ch = 0;

			/* Get the mouse protocol response */
			do {
				FD_ZERO(&fdset);
				FD_SET(fd, &fdset);
				tv.tv_sec = 1;
				tv.tv_usec = 0;
				if ( select(fd+1, &fdset, 0, 0, &tv) < 1 ) {
					break;
				}
			} while ( (read(fd, &ch, sizeof (ch)) == sizeof (ch)) &&
			          ((ch == 0xFA) || (ch == 0xAA)) );

			/* Experimental values (Logitech wheelmouse) */
#ifdef DEBUG_MOUSE
fprintf(stderr, "Last mouse mode: 0x%x\n", ch);
#endif
			if ( (ch == 3) || (ch == 4) ) {
				imps2 = 1;
			}
		}
	}
	return imps2;
}

int FB_OpenMouse(_THIS)
{
	int i;
	const char *mousedev;
	const char *mousedrv;

	mousedrv = getenv("SDL_MOUSEDRV");
	mousedev = getenv("SDL_MOUSEDEV");
	mouse_fd = -1;

	/* ELO TOUCHSCREEN SUPPORT */

	if( (mousedrv != NULL) && (strcmp(mousedrv, "ELO") == 0) ) {
		mouse_fd = open(mousedev, O_RDWR);
		if ( mouse_fd >= 0 ) {
			if(eloInitController(mouse_fd)) {
#ifdef DEBUG_MOUSE
fprintf(stderr, "Using ELO touchscreen\n");
#endif
   				mouse_drv = MOUSE_ELO;
			}

		}
		else if ( mouse_fd < 0 ) {
			mouse_drv = MOUSE_NONE;
		}

		return(mouse_fd);
	}

	/* STD MICE */

	if ( mousedev == NULL ) {
		/* FIXME someday... allow multiple mice in this driver */
		static const char * const ps2mice[] = {
		    "/dev/input/mice", "/dev/usbmouse", "/dev/psaux", NULL
		};
		/* First try to use GPM in repeater mode */
		if ( mouse_fd < 0 ) {
			if ( gpm_available() ) {
				mouse_fd = open(GPM_NODE_FIFO, O_RDONLY, 0);
				if ( mouse_fd >= 0 ) {
#ifdef DEBUG_MOUSE
fprintf(stderr, "Using GPM mouse\n");
#endif
					mouse_drv = MOUSE_MSC;
				}
			}
		}
		/* Now try to use a modern PS/2 mouse */
		for ( i=0; (mouse_fd < 0) && ps2mice[i]; ++i ) {
			mouse_fd = open(ps2mice[i], O_RDWR, 0);
			if (mouse_fd < 0) {
				mouse_fd = open(ps2mice[i], O_RDONLY, 0);
			}
			if (mouse_fd >= 0) {
				/* rcg06112001 Attempt to set IMPS/2 mode */
				if ( i == 0 ) {
					set_imps2_mode(mouse_fd);
				}
				if (detect_imps2(mouse_fd)) {
#ifdef DEBUG_MOUSE
fprintf(stderr, "Using IMPS2 mouse\n");
#endif
					mouse_drv = MOUSE_IMPS2;
				} else {
#ifdef DEBUG_MOUSE
fprintf(stderr, "Using PS2 mouse\n");
#endif
					mouse_drv = MOUSE_PS2;
				}
			}
		}
		/* Next try to use a PPC ADB port mouse */
		if ( mouse_fd < 0 ) {
			mouse_fd = open("/dev/adbmouse", O_RDONLY, 0);
			if ( mouse_fd >= 0 ) {
#ifdef DEBUG_MOUSE
fprintf(stderr, "Using ADB mouse\n");
#endif
				mouse_drv = MOUSE_BM;
			}
		}
	}
	/* Default to a serial Microsoft mouse */
	if ( mouse_fd < 0 ) {
		if ( mousedev == NULL ) {
			mousedev = "/dev/mouse";
		}
		mouse_fd = open(mousedev, O_RDONLY, 0);
		if ( mouse_fd >= 0 ) {
			struct termios mouse_termios;

			/* Set the sampling speed to 1200 baud */
			tcgetattr(mouse_fd, &mouse_termios);
			mouse_termios.c_iflag = IGNBRK | IGNPAR;
			mouse_termios.c_oflag = 0;
			mouse_termios.c_lflag = 0;
			mouse_termios.c_line = 0;
			mouse_termios.c_cc[VTIME] = 0;
			mouse_termios.c_cc[VMIN] = 1;
			mouse_termios.c_cflag = CREAD | CLOCAL | HUPCL;
			mouse_termios.c_cflag |= CS8;
			mouse_termios.c_cflag |= B1200;
			tcsetattr(mouse_fd, TCSAFLUSH, &mouse_termios);
#ifdef DEBUG_MOUSE
fprintf(stderr, "Using Microsoft mouse on %s\n", mousedev);
#endif
			mouse_drv = MOUSE_MS;
		}
	}
	if ( mouse_fd < 0 ) {
		mouse_drv = MOUSE_NONE;
	}
	return(mouse_fd);
}

static int posted = 0;

void FB_vgamousecallback(int button, int relative, int dx, int dy)
{
	int button_1, button_3;
	int button_state;
	int state_changed;
	int i;
	Uint8 state;

	if ( dx || dy ) {
		posted += SDL_PrivateMouseMotion(0, relative, dx, dy);
	}

	/* Swap button 1 and 3 */
	button_1 = (button & 0x04) >> 2;
	button_3 = (button & 0x01) << 2;
	button &= ~0x05;
	button |= (button_1|button_3);

	/* See what changed */
	button_state = SDL_GetMouseState(NULL, NULL);
	state_changed = button_state ^ button;
	for ( i=0; i<8; ++i ) {
		if ( state_changed & (1<<i) ) {
			if ( button & (1<<i) ) {
				state = SDL_PRESSED;
			} else {
				state = SDL_RELEASED;
			}
			posted += SDL_PrivateMouseButton(state, i+1, 0, 0);
		}
	}
}

/* For now, use MSC, PS/2, and MS protocols
   Driver adapted from the SVGAlib mouse driver code (taken from gpm, etc.)
 */
#if 0
static void handle_mouse(_THIS)
{
	static int start = 0;
	static unsigned char mousebuf[BUFSIZ];
	static int relative = 1;

	int i, nread;
	int button = 0;
	int dx = 0, dy = 0;
	int packetsize = 0;
	int realx, realy;
	
	/* Figure out the mouse packet size */
	switch (mouse_drv) {
		case MOUSE_NONE:
			/* Ack! */
			read(mouse_fd, mousebuf, BUFSIZ);
			return;
		case MOUSE_MSC:
			packetsize = 5;
			break;
		case MOUSE_IMPS2:
			packetsize = 4;
			break;
		case MOUSE_PS2:
		case MOUSE_MS:
		case MOUSE_BM:
			packetsize = 3;
			break;
		case MOUSE_ELO:
			packetsize = ELO_PACKET_SIZE;
			relative = 0;
			break;
		case NUM_MOUSE_DRVS:
			/* Uh oh.. */
			packetsize = 0;
			break;
	}

	/* Special handling for the quite sensitive ELO controller */
	if (mouse_drv == MOUSE_ELO) {
	
	    /* try to read the next packet */
	    if(eloReadPosition(this, mouse_fd, &dx, &dy, &button, &realx, &realy)) {
		button = (button & 0x01) << 2;
    		FB_vgamousecallback(button, relative, dx, dy);
	    }
	    
	    return;
	}
	
	/* Read as many packets as possible */
	nread = read(mouse_fd, &mousebuf[start], BUFSIZ-start);
	if ( nread < 0 ) {
		return;
	}
	nread += start;
#ifdef DEBUG_MOUSE
	fprintf(stderr, "Read %d bytes from mouse, start = %d\n", nread, start);
#endif
	for ( i=0; i<(nread-(packetsize-1)); i += packetsize ) {
		switch (mouse_drv) {
			case MOUSE_NONE:
				break;
			case MOUSE_MSC:
				/* MSC protocol has 0x80 in high byte */
				if ( (mousebuf[i] & 0xF8) != 0x80 ) {
					/* Go to next byte */
					i -= (packetsize-1);
					continue;
				}
				/* Get current mouse state */
				button = (~mousebuf[i]) & 0x07;
				dx =   (signed char)(mousebuf[i+1]) +
				       (signed char)(mousebuf[i+3]);
				dy = -((signed char)(mousebuf[i+2]) +
				       (signed char)(mousebuf[i+4]));
				break;
			case MOUSE_PS2:
				/* PS/2 protocol has nothing in high byte */
				if ( (mousebuf[i] & 0xC0) != 0 ) {
					/* Go to next byte */
					i -= (packetsize-1);
					continue;
				}
				/* Get current mouse state */
				button = (mousebuf[i] & 0x04) >> 1 | /*Middle*/
		  			 (mousebuf[i] & 0x02) >> 1 | /*Right*/
		  			 (mousebuf[i] & 0x01) << 2;  /*Left*/
		  		dx = (mousebuf[i] & 0x10) ?
		  		      mousebuf[i+1] - 256 : mousebuf[i+1];
		  		dy = (mousebuf[i] & 0x20) ?
		  		      -(mousebuf[i+2] - 256) : -mousebuf[i+2];
				break;
			case MOUSE_IMPS2:
				/* Get current mouse state */
				button = (mousebuf[i] & 0x04) >> 1 | /*Middle*/
		  			 (mousebuf[i] & 0x02) >> 1 | /*Right*/
		  			 (mousebuf[i] & 0x01) << 2 | /*Left*/
		  			 (mousebuf[i] & 0x40) >> 3 | /* 4 */
		  			 (mousebuf[i] & 0x80) >> 3;  /* 5 */
		  		dx = (mousebuf[i] & 0x10) ?
		  		      mousebuf[i+1] - 256 : mousebuf[i+1];
		  		dy = (mousebuf[i] & 0x20) ?
		  		      -(mousebuf[i+2] - 256) : -mousebuf[i+2];
				switch (mousebuf[i+3]&0x0F) {
				    case 0x0E: /* DX = +1 */
				    case 0x02: /* DX = -1 */
					break;
				    case 0x0F: /* DY = +1 (map button 4) */
                                       FB_vgamousecallback(button | (1<<3),
                                                           1, 0, 0);
					break;
				    case 0x01: /* DY = -1 (map button 5) */
                                       FB_vgamousecallback(button | (1<<4),
                                                           1, 0, 0);
					break;
				}
				break;
			case MOUSE_MS:
				/* Microsoft protocol has 0x40 in high byte */
				if ( (mousebuf[i] & 0x40) != 0x40 ) {
					/* Go to next byte */
					i -= (packetsize-1);
					continue;
				}
				/* Get current mouse state */
				button = ((mousebuf[i] & 0x20) >> 3) |
				         ((mousebuf[i] & 0x10) >> 4);
				dx = (signed char)(((mousebuf[i] & 0x03) << 6) |
				                   (mousebuf[i + 1] & 0x3F));
				dy = (signed char)(((mousebuf[i] & 0x0C) << 4) |
				                    (mousebuf[i + 2] & 0x3F));
				break;
			case MOUSE_BM:
				/* BusMouse protocol has 0xF8 in high byte */
				if ( (mousebuf[i] & 0xF8) != 0x80 ) {
					/* Go to next byte */
					i -= (packetsize-1);
					continue;
				}
				/* Get current mouse state */
				button = (~mousebuf[i]) & 0x07;
				dx =  (signed char)mousebuf[i+1];
				dy = -(signed char)mousebuf[i+2];
				break;
			/*
			case MOUSE_ELO:
				if ( mousebuf[i] != ELO_START_BYTE ) {
					i -= (packetsize-1);
					continue;
				}

				if(!eloParsePacket(&(mousebuf[i]), &dx, &dy, &button)) {
					i -= (packetsize-1);
					continue;
				}
				
				button = (button & 0x01) << 2;

				eloConvertXY(this, &dx, &dy);
				break;
			*/

			case MOUSE_ELO:
			case NUM_MOUSE_DRVS:
				/* Uh oh.. */
				dx = 0;
				dy = 0;
				break;
		}
		FB_vgamousecallback(button, relative, dx, dy);
	}
	if ( i < nread ) {
		memcpy(mousebuf, &mousebuf[i], (nread-i));
		start = (nread-i);
	} else {
		start = 0;
	}
	return;
}
#endif

/* Handle switching to another VC, returns when our VC is back.
   This isn't necessarily the best solution.  For SDL 1.3 we need
   a way of notifying the application when we lose access to the
   video hardware and when we regain it.
 */
#if 0
static void switch_vt(_THIS, unsigned short which)
{
	struct vt_stat vtstate;
	unsigned short current;
	SDL_Surface *screen;
	__u16 saved_pal[3*256];
	Uint32 screen_arealen;
	Uint8 *screen_contents;

	/* Figure out whether or not we're switching to a new console */
	if ( (ioctl(keyboard_fd, VT_GETSTATE, &vtstate) < 0) ||
	     (which == vtstate.v_active) ) {
		return;
	}
	current = vtstate.v_active;

	/* Save the contents of the screen, and go to text mode */
	SDL_mutexP(hw_lock);
	wait_idle(this);
	screen = SDL_VideoSurface;
	screen_arealen = (screen->h*screen->pitch);
	screen_contents = (Uint8 *)malloc(screen_arealen);
	if ( screen_contents ) {
		memcpy(screen_contents, screen->pixels, screen_arealen);
	}
	FB_SavePaletteTo(this, 256, saved_pal);
	ioctl(keyboard_fd, KDSETMODE, KD_TEXT);

	/* New console, switch to it */
	if ( ioctl(keyboard_fd, VT_ACTIVATE, which) == 0 ) {
		/* Wait for our console to be activated again */
		ioctl(keyboard_fd, VT_WAITACTIVE, which);
		while ( ioctl(keyboard_fd, VT_WAITACTIVE, current) < 0 ) {
			if ( (errno != EINTR) && (errno != EAGAIN) ) {
				/* Unknown VT error - cancel this */
				break;
			}
			SDL_Delay(500);
		}
	}

	/* Restore graphics mode and the contents of the screen */
	ioctl(keyboard_fd, KDSETMODE, KD_GRAPHICS);
	FB_RestorePaletteFrom(this, 256, saved_pal);
	if ( screen_contents ) {
		memcpy(screen->pixels, screen_contents, screen_arealen);
		free(screen_contents);
	}
	SDL_mutexV(hw_lock);
}
#endif

static void handle_keyboard(_THIS)
{
//#define	DEBUG_KEYBOARD
#define		IR_REPEAT_DELAY_FIRST		300
#define		IR_REPEAT_DELAY_NEXT		150

	unsigned char		keybuf; // keybuf[2];
	//int				i,
	int					nread;
	int					pressed;
	SDL_keysym			keysym;

	static int			version		= 0;
	static int			lastKeySym	= 0;
	static int			lastEvent	= 0;
	static SDL_keysym	lastSDLKey	= { 0, SDLK_UNKNOWN, KMOD_NONE, 0, 0 };
	static Uint32		waitRepeat	= IR_REPEAT_DELAY_FIRST;
	
	static int			capslock = 0;
	static int			usbkbcount = 0;
	int					usbkb;
	
	if (++usbkbcount > 200) {
		usbkbcount = 0;
		FB_OpenKeyboard(this);
	}

	// Front panel
	
	nread = read (keyboard_fd, &keybuf, 1 /*2*/);

	//for (i = 0; i < nread; ++i)
	if (nread == 1)
	{
		// 0xff end marker
		if (keybuf == 0xff)
		{
			// Outstanding SDL_RELEASED for IR keys?
			if (lastSDLKey.scancode && (lastSDLKey.when + waitRepeat < SDL_GetTicks ()))
			{
				#ifdef DEBUG_KEYBOARD
					fprintf(stderr, "0x%08X delayed sdlkey: %s\n", SDL_GetTicks (), SDL_GetKeyName (lastSDLKey.sym));
				#endif
				posted += SDL_PrivateKeyboard (SDL_RELEASED, &lastSDLKey);
				
				// Clear old key event
				memset (&lastSDLKey, 0, sizeof (SDL_keysym));
				// Time til first event is repeated is longer than later repeat intervals
				waitRepeat = IR_REPEAT_DELAY_FIRST;
			}
			//continue;
		}
		// 0x20 version info
		else if (keybuf == 0x20)
		{
			#ifdef DEBUG_KEYBOARD
				fprintf(stderr, "0x%08X version key: 0x%02X\n", SDL_GetTicks (), keybuf);
			#endif
			version = keybuf;
		}
		else
		{
			// Original and mapped scancode
			keysym.scancode = keybuf;
			keysym.sym		= gKeyCodes[keysym.scancode].sdlKey;
			/* DF timestamp, should be better retrieved from keyboard itself */
			keysym.when		= SDL_GetTicks ();
			pressed			= gKeyCodes[keysym.scancode].keyPressed;

			#ifdef DEBUG_KEYBOARD
				fprintf(stderr, "0x%08X keybuf: 0x%02X / %d -> %d\n", SDL_GetTicks (), keybuf, keybuf, keysym.sym);
			#endif
			
			// Special handling for inner wheel
			if ((keysym.scancode == 13) || (keysym.scancode == 14))
			{
				// Ignore too "fast" events (prelling)
				if (keysym.when - lastEvent < 40)
				{
					#ifdef DEBUG_KEYBOARD
						fprintf(stderr, "0x%08X ignore sdlkey: %s at/last: %d/%d\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym), keysym.when, lastEvent);
					#endif
					//continue;
				}
				else
				{
					lastEvent = keysym.when;
				
					#ifdef DEBUG_KEYBOARD
						fprintf(stderr, "0x%08X pressed sdlkey:  %s at %d\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym), keysym.when);
					#endif
					posted += SDL_PrivateKeyboard (SDL_PRESSED, &keysym);
					#ifdef DEBUG_KEYBOARD
						fprintf(stderr, "0x%08X released sdlkey: %s at %d\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym), keysym.when);
					#endif
					posted += SDL_PrivateKeyboard (SDL_RELEASED, &keysym);
				}
			}
			// Special handling for outer wheel
			else if ((keysym.scancode >= 16) && (keysym.scancode <= 30))
			{
				switch (keysym.scancode)
				{
					// Released position
					case 16:
						if (lastKeySym)
						{
							keysym.sym	= lastKeySym; 
							lastKeySym	= 0;					// release only once
							pressed		= SDL_RELEASED;
						}
						break;
					
					// Outer wheel right positions
					case 17:	// case 18 .. 23 unused
						// Last position was from other side of wheel, release it first
						if (lastKeySym && lastKeySym != SDLK_RIGHT)
						{
							keysym.sym = SDLK_LEFT;
							#ifdef DEBUG_KEYBOARD
								fprintf(stderr, "0x%08X simulate missing released sdlkey: %s\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym));
							#endif
							posted += SDL_PrivateKeyboard (SDL_RELEASED, &keysym);
						}
						keysym.sym	= SDLK_RIGHT;
						lastKeySym	= keysym.sym;
						pressed		= SDL_PRESSED;
						break;

					// Outer wheel left positions
					case 24:	// case 25 .. 30 unused
						// Last position was from other side of wheel, release it first
						if (lastKeySym && lastKeySym != SDLK_LEFT)
						{
							keysym.sym = SDLK_RIGHT;
							#ifdef DEBUG_KEYBOARD
								fprintf(stderr, "0x%08X simulate missing released sdlkey: %s\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym));
							#endif
							posted += SDL_PrivateKeyboard (SDL_RELEASED, &keysym);
						}
						keysym.sym	= SDLK_LEFT;
						lastKeySym	= keysym.sym;
						pressed		= SDL_PRESSED;
						break;

					default:
						keysym.sym	= SDLK_UNKNOWN;
						break;
				}
				if (keysym.sym != SDLK_UNKNOWN)
				{
					#ifdef DEBUG_KEYBOARD
						if (pressed)
							fprintf(stderr, "0x%08X pressed sdlkey:  %s\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym));
						else
							fprintf(stderr, "0x%08X released sdlkey: %s\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym));
					#endif
					posted += SDL_PrivateKeyboard (pressed, &keysym);
				}
			}
			// All others including IR control
			else
			{
				// Map scancode to SDL key
				if (keysym.sym != SDLK_UNKNOWN)
				{
					#ifdef DEBUG_KEYBOARD
						if (pressed)
							fprintf(stderr, "0x%08X pressed sdlkey:  %s\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym));
						else
							fprintf(stderr, "0x%08X released sdlkey: %s\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym));
					#endif
					posted += SDL_PrivateKeyboard (pressed, &keysym);
					
					// If key will not be released (all IR), release code must be posted
					if (gKeyCodes[keysym.scancode].simReleased)
					{
						// Newer versions can repeat events, so release code is posted via time out
						if ((version >= 0x20) && gKeyCodes[keysym.scancode].canRepeat)
						{
							#ifdef DEBUG_KEYBOARD
								fprintf(stderr, "0x%08X delay released sdlkey: %s\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym));
							#endif
							// Event is repeated, intervals are smaller now
							if (lastSDLKey.scancode == keysym.scancode)
								waitRepeat = IR_REPEAT_DELAY_NEXT;
							lastSDLKey = keysym;
						}
						// Post release code immediately
						else
						{
							#ifdef DEBUG_KEYBOARD
								fprintf(stderr, "0x%08X released sdlkey: %s\n", SDL_GetTicks (), SDL_GetKeyName (keysym.sym));
							#endif
							posted += SDL_PrivateKeyboard (SDL_RELEASED, &keysym);
						}
					}
				}
				#ifdef DEBUG_KEYBOARD
				else
					fprintf(stderr, "0x%08X unmapped sdlkey\n", SDL_GetTicks ());
				#endif
			}
		}
	}
	
	// USB keyboard

	for ( usbkb = 0; usbkb < 2; ++usbkb ) {
		int ret;
		struct input_event ev;

		if (usbkeyboard_fd[usbkb] < 0) continue;

		ret = read(usbkeyboard_fd[usbkb], &ev, sizeof(ev));
		if (ret == sizeof(ev))
		{
			#ifdef DEBUG_KEYBOARD
				fprintf(stderr, "USB Keyboard event code: %d\n", ev.code);
			#endif
			
			switch (ev.value)
			{
				case 0: // release
				case 1: // press
					switch (ev.code)
					{
						case 172: // Homepage
							keysym.sym = SDLK_HOMEPAGE;
							break;
						case 155: // Mail
							keysym.sym = SDLK_MAIL;
							break;
						case 144: // File
							keysym.sym = SDLK_FILE;
							break;
						case 143: // WakeUp
							keysym.sym = SDLK_SLEEP;
							break;
						case 116: // Power
							keysym.sym = SDLK_HALT;
							break;

						case 58: { // caps lock
							if (ev.value) {
								capslock ^= 1;
								struct input_event evl;
								evl.type = EV_LED;
								evl.code = LED_CAPSL;
								evl.value = capslock;
								write(usbkeyboard_fd[usbkb], &evl, sizeof(evl));
							}
							// no break
						}

						default:
							if (ev.code <= 127) {
								keysym.sym = keymap[ev.code];
							}
							else {
								keysym.sym = SDLK_UNKNOWN;
							}
							break;
					}
						
					keysym.scancode = ev.code;
					/* DF timestamp, should be better retrieved from keyboard itself */
					keysym.when		= SDL_GetTicks ();
					#ifdef DEBUG_KEYBOARD
						fprintf(stderr, "SDL scancode: %d keysym: %d\n", keysym.scancode, keysym.sym);
					#endif
					posted += SDL_PrivateKeyboard (ev.value ? SDL_PRESSED : SDL_RELEASED, &keysym);
					break;
				default:
					break;
			}
		}
		else {
			if (ret < 0 && errno != EAGAIN) {
				#ifdef DEBUG_KEYBOARD
					fprintf(stderr, "Read from USB kbd %d \'%s\'\n", ret, strerror(errno));
				#endif
				close(usbkeyboard_fd[usbkb]);
				usbkeyboard_fd[usbkb] = -1;
				break;
			}
		}
	}
}

void FB_PumpEvents(_THIS)
{
#if 0
	fd_set fdset;
	int max_fd;
	static struct timeval zero;
#endif

	do {
		posted = 0;

#if 0
		FD_ZERO(&fdset);
		max_fd = 0;
		if ( keyboard_fd >= 0 ) {
			FD_SET(keyboard_fd, &fdset);
			if ( max_fd < keyboard_fd ) {
				max_fd = keyboard_fd;
			}
		}
		if ( mouse_fd >= 0 ) {
			FD_SET(mouse_fd, &fdset);
			if ( max_fd < mouse_fd ) {
				max_fd = mouse_fd;
			}
		}
		if ( select(max_fd+1, &fdset, NULL, NULL, &zero) > 0 ) {
			if ( keyboard_fd >= 0 ) {
				if ( FD_ISSET(keyboard_fd, &fdset) ) {
					handle_keyboard(this);
				}
			}
			if ( mouse_fd >= 0 ) {
				if ( FD_ISSET(mouse_fd, &fdset) ) {
					handle_mouse(this);
				}
			}
		}
#endif
	//	SDL_Delay(50); /*DF*/
		handle_keyboard(this);

	} while ( posted );
}

#if 0
void FB_InitOSKeymap(_THIS)
{
	int i;

	/* Initialize the Linux key translation table */

	/* First get the ascii keys and others not well handled */
	for (i=0; i<SDL_TABLESIZE(keymap); ++i) {
	  switch(i) {
	  /* These aren't handled by the x86 kernel keymapping (?) */
	  case SCANCODE_PRINTSCREEN:
	    keymap[i] = SDLK_PRINT;
	    break;
	  case SCANCODE_BREAK:
	    keymap[i] = SDLK_BREAK;
	    break;
	  case SCANCODE_BREAK_ALTERNATIVE:
	    keymap[i] = SDLK_PAUSE;
	    break;
	  case SCANCODE_LEFTSHIFT:
	    keymap[i] = SDLK_LSHIFT;
	    break;
	  case SCANCODE_RIGHTSHIFT:
	    keymap[i] = SDLK_RSHIFT;
	    break;
	  case SCANCODE_LEFTCONTROL:
	    keymap[i] = SDLK_LCTRL;
	    break;
	  case SCANCODE_RIGHTCONTROL:
	    keymap[i] = SDLK_RCTRL;
	    break;
	  case SCANCODE_RIGHTWIN:
	    keymap[i] = SDLK_RSUPER;
	    break;
	  case SCANCODE_LEFTWIN:
	    keymap[i] = SDLK_LSUPER;
	    break;
	  case 127:
	    keymap[i] = SDLK_MENU;
	    break;
	  /* this should take care of all standard ascii keys */
	  default:
	    keymap[i] = KVAL(vga_keymap[0][i]);
	    break;
          }
	}
	for (i=0; i<SDL_TABLESIZE(keymap); ++i) {
	  switch(keymap_temp[i]) {
	    case K_F1:  keymap[i] = SDLK_F1;  break;
	    case K_F2:  keymap[i] = SDLK_F2;  break;
	    case K_F3:  keymap[i] = SDLK_F3;  break;
	    case K_F4:  keymap[i] = SDLK_F4;  break;
	    case K_F5:  keymap[i] = SDLK_F5;  break;
	    case K_F6:  keymap[i] = SDLK_F6;  break;
	    case K_F7:  keymap[i] = SDLK_F7;  break;
	    case K_F8:  keymap[i] = SDLK_F8;  break;
	    case K_F9:  keymap[i] = SDLK_F9;  break;
	    case K_F10: keymap[i] = SDLK_F10; break;
	    case K_F11: keymap[i] = SDLK_F11; break;
	    case K_F12: keymap[i] = SDLK_F12; break;

	    case K_DOWN:  keymap[i] = SDLK_DOWN;  break;
	    case K_LEFT:  keymap[i] = SDLK_LEFT;  break;
	    case K_RIGHT: keymap[i] = SDLK_RIGHT; break;
	    case K_UP:    keymap[i] = SDLK_UP;    break;

	    case K_P0:     keymap[i] = SDLK_KP0; break;
	    case K_P1:     keymap[i] = SDLK_KP1; break;
	    case K_P2:     keymap[i] = SDLK_KP2; break;
	    case K_P3:     keymap[i] = SDLK_KP3; break;
	    case K_P4:     keymap[i] = SDLK_KP4; break;
	    case K_P5:     keymap[i] = SDLK_KP5; break;
	    case K_P6:     keymap[i] = SDLK_KP6; break;
	    case K_P7:     keymap[i] = SDLK_KP7; break;
	    case K_P8:     keymap[i] = SDLK_KP8; break;
	    case K_P9:     keymap[i] = SDLK_KP9; break;
	    case K_PPLUS:  keymap[i] = SDLK_KP_PLUS; break;
	    case K_PMINUS: keymap[i] = SDLK_KP_MINUS; break;
	    case K_PSTAR:  keymap[i] = SDLK_KP_MULTIPLY; break;
	    case K_PSLASH: keymap[i] = SDLK_KP_DIVIDE; break;
	    case K_PENTER: keymap[i] = SDLK_KP_ENTER; break;
	    case K_PDOT:   keymap[i] = SDLK_KP_PERIOD; break;

	    case K_SHIFT:  if ( keymap[i] != SDLK_RSHIFT )
	                     keymap[i] = SDLK_LSHIFT;
	                   break;
	    case K_SHIFTL: keymap[i] = SDLK_LSHIFT; break;
	    case K_SHIFTR: keymap[i] = SDLK_RSHIFT; break;
	    case K_CTRL:  if ( keymap[i] != SDLK_RCTRL )
	                     keymap[i] = SDLK_LCTRL;
	                   break;
	    case K_CTRLL:  keymap[i] = SDLK_LCTRL;  break;
	    case K_CTRLR:  keymap[i] = SDLK_RCTRL;  break;
	    case K_ALT:    keymap[i] = SDLK_LALT;   break;
	    case K_ALTGR:  keymap[i] = SDLK_RALT;   break;

	    case K_INSERT: keymap[i] = SDLK_INSERT;   break;
	    case K_REMOVE: keymap[i] = SDLK_DELETE;   break;
	    case K_PGUP:   keymap[i] = SDLK_PAGEUP;   break;
	    case K_PGDN:   keymap[i] = SDLK_PAGEDOWN; break;
	    case K_FIND:   keymap[i] = SDLK_HOME;     break;
	    case K_SELECT: keymap[i] = SDLK_END;      break;

	    case K_NUM:  keymap[i] = SDLK_NUMLOCK;   break;
	    case K_CAPS: keymap[i] = SDLK_CAPSLOCK;  break;

	    case K_F13:   keymap[i] = SDLK_PRINT;     break;
	    case K_HOLD:  keymap[i] = SDLK_SCROLLOCK; break;
	    case K_PAUSE: keymap[i] = SDLK_PAUSE;     break;

	    case 127: keymap[i] = SDLK_BACKSPACE; break;
	     
	    default: break;
	  }
	}
}
#endif

void FB_InitOSKeymap(_THIS) // MG: taken from SVGA_InitOSKeymap()
{
	int i;

	for ( i=0; i<SDL_TABLESIZE(keymap); ++i )
		keymap[i] = SDLK_UNKNOWN;

	keymap[SCANCODE_ESCAPE]			= SDLK_ESCAPE;
	keymap[SCANCODE_1]				= SDLK_1;
	keymap[SCANCODE_2]				= SDLK_2;
	keymap[SCANCODE_3]				= SDLK_3;
	keymap[SCANCODE_4]				= SDLK_4;
	keymap[SCANCODE_5]				= SDLK_5;
	keymap[SCANCODE_6]				= SDLK_6;
	keymap[SCANCODE_7]				= SDLK_7;
	keymap[SCANCODE_8]				= SDLK_8;
	keymap[SCANCODE_9]				= SDLK_9;
	keymap[SCANCODE_0]				= SDLK_0;
	keymap[SCANCODE_MINUS]			= SDLK_WORLD_1; // sz
	keymap[SCANCODE_EQUAL]			= SDLK_QUOTE;
	keymap[SCANCODE_BACKSPACE]		= SDLK_BACKSPACE;
	keymap[SCANCODE_TAB]			= SDLK_TAB;
	keymap[SCANCODE_Q]				= SDLK_q;
	keymap[SCANCODE_W]				= SDLK_w;
	keymap[SCANCODE_E]				= SDLK_e;
	keymap[SCANCODE_R]				= SDLK_r;
	keymap[SCANCODE_T]				= SDLK_t;
	keymap[SCANCODE_Y]				= SDLK_z;
	keymap[SCANCODE_U]				= SDLK_u;
	keymap[SCANCODE_I]				= SDLK_i;
	keymap[SCANCODE_O]				= SDLK_o;
	keymap[SCANCODE_P]				= SDLK_p;
	keymap[SCANCODE_BRACKET_LEFT]	= SDLK_WORLD_2; // ue
	keymap[SCANCODE_BRACKET_RIGHT]	= SDLK_PLUS;
	keymap[SCANCODE_ENTER]			= SDLK_RETURN;
	keymap[SCANCODE_LEFTCONTROL]	= SDLK_LCTRL;
	keymap[SCANCODE_A]				= SDLK_a;
	keymap[SCANCODE_S]				= SDLK_s;
	keymap[SCANCODE_D]				= SDLK_d;
	keymap[SCANCODE_F]				= SDLK_f;
	keymap[SCANCODE_G]				= SDLK_g;
	keymap[SCANCODE_H]				= SDLK_h;
	keymap[SCANCODE_J]				= SDLK_j;
	keymap[SCANCODE_K]				= SDLK_k;
	keymap[SCANCODE_L]				= SDLK_l;
	keymap[SCANCODE_SEMICOLON]		= SDLK_WORLD_4; // oe
	keymap[SCANCODE_APOSTROPHE]		= SDLK_WORLD_3; // ae
	keymap[SCANCODE_GRAVE]			= SDLK_BACKQUOTE;
	keymap[SCANCODE_LEFTSHIFT]		= SDLK_LSHIFT;
	keymap[SCANCODE_BACKSLASH]		= SDLK_BACKSLASH;
	keymap[SCANCODE_Z]				= SDLK_y;
	keymap[SCANCODE_X]				= SDLK_x;
	keymap[SCANCODE_C]				= SDLK_c;
	keymap[SCANCODE_V]				= SDLK_v;
	keymap[SCANCODE_B]				= SDLK_b;
	keymap[SCANCODE_N]				= SDLK_n;
	keymap[SCANCODE_M]				= SDLK_m;
	keymap[SCANCODE_COMMA]			= SDLK_COMMA;
	keymap[SCANCODE_PERIOD]			= SDLK_PERIOD;
	keymap[SCANCODE_SLASH]			= SDLK_MINUS;
	keymap[SCANCODE_RIGHTSHIFT]		= SDLK_RSHIFT;
	keymap[SCANCODE_KEYPADMULTIPLY]	= SDLK_KP_MULTIPLY;
	keymap[SCANCODE_LEFTALT]		= SDLK_LALT;
	keymap[SCANCODE_SPACE]			= SDLK_SPACE;
	keymap[SCANCODE_CAPSLOCK]		= SDLK_CAPSLOCK;
	keymap[SCANCODE_F1]				= SDLK_F1;
	keymap[SCANCODE_F2]				= SDLK_F2;
	keymap[SCANCODE_F3]				= SDLK_F3;
	keymap[SCANCODE_F4]				= SDLK_F4;
	keymap[SCANCODE_F5]				= SDLK_F5;
	keymap[SCANCODE_F6]				= SDLK_F6;
	keymap[SCANCODE_F7]				= SDLK_F7;
	keymap[SCANCODE_F8]				= SDLK_F8;
	keymap[SCANCODE_F9]				= SDLK_F9;
	keymap[SCANCODE_F10]			= SDLK_F10;
	keymap[SCANCODE_NUMLOCK]		= SDLK_NUMLOCK;
	keymap[SCANCODE_SCROLLLOCK]		= SDLK_SCROLLOCK;
	keymap[SCANCODE_KEYPAD7]		= SDLK_KP7;
	keymap[SCANCODE_CURSORUPLEFT]	= SDLK_KP7;
	keymap[SCANCODE_KEYPAD8]		= SDLK_KP8;
	keymap[SCANCODE_CURSORUP]		= SDLK_KP8;
	keymap[SCANCODE_KEYPAD9]		= SDLK_KP9;
	keymap[SCANCODE_CURSORUPRIGHT]	= SDLK_KP9;
	keymap[SCANCODE_KEYPADMINUS]	= SDLK_KP_MINUS;
	keymap[SCANCODE_KEYPAD4]		= SDLK_KP4;
	keymap[SCANCODE_CURSORLEFT]		= SDLK_KP4;
	keymap[SCANCODE_KEYPAD5]		= SDLK_KP5;
	keymap[SCANCODE_KEYPAD6]		= SDLK_KP6;
	keymap[SCANCODE_CURSORRIGHT]	= SDLK_KP6;
	keymap[SCANCODE_KEYPADPLUS]		= SDLK_KP_PLUS;
	keymap[SCANCODE_KEYPAD1]		= SDLK_KP1;
	keymap[SCANCODE_CURSORDOWNLEFT]	= SDLK_KP1;
	keymap[SCANCODE_KEYPAD2]		= SDLK_KP2;
	keymap[SCANCODE_CURSORDOWN]		= SDLK_KP2;
	keymap[SCANCODE_KEYPAD3]		= SDLK_KP3;
	keymap[SCANCODE_CURSORDOWNRIGHT]= SDLK_KP3;
	keymap[SCANCODE_KEYPAD0]		= SDLK_KP0;
	keymap[SCANCODE_KEYPADPERIOD]	= SDLK_KP_PERIOD;
	keymap[84]						= SDLK_HASH;
	keymap[SCANCODE_LESS]			= SDLK_LESS;
	keymap[SCANCODE_F11]			= SDLK_F11;
	keymap[SCANCODE_F12]			= SDLK_F12;
	keymap[SCANCODE_KEYPADENTER]	= SDLK_KP_ENTER;
	keymap[SCANCODE_RIGHTCONTROL]	= SDLK_RCTRL;
	keymap[SCANCODE_CONTROL]		= SDLK_RCTRL;
	keymap[SCANCODE_KEYPADDIVIDE]	= SDLK_KP_DIVIDE;
	keymap[SCANCODE_PRINTSCREEN]	= SDLK_PRINT;
	keymap[SCANCODE_RIGHTALT]		= SDLK_RALT;
	keymap[SCANCODE_BREAK]			= SDLK_BREAK;
	keymap[SCANCODE_BREAK_ALTERNATIVE]= SDLK_UNKNOWN;
	keymap[SCANCODE_HOME]			= SDLK_HOME;
	keymap[SCANCODE_CURSORBLOCKUP]	= SDLK_UP;
	keymap[SCANCODE_PAGEUP]			= SDLK_PAGEUP;
	keymap[SCANCODE_CURSORBLOCKLEFT]= SDLK_LEFT;
	keymap[SCANCODE_CURSORBLOCKRIGHT]= SDLK_RIGHT;
	keymap[SCANCODE_END]			= SDLK_END;
	keymap[SCANCODE_CURSORBLOCKDOWN]= SDLK_DOWN;
	keymap[SCANCODE_PAGEDOWN]		= SDLK_PAGEDOWN;
	keymap[SCANCODE_INSERT]			= SDLK_INSERT;
	keymap[SCANCODE_REMOVE]			= SDLK_DELETE;
	keymap[119]						= SDLK_PAUSE;
	keymap[SCANCODE_RIGHTWIN]		= SDLK_RSUPER;
	keymap[SCANCODE_LEFTWIN]		= SDLK_LSUPER;
	keymap[127]						= SDLK_MENU;
}

#if 0
static SDL_keysym *TranslateKey(int scancode, SDL_keysym *keysym)
{
	/* Set the keysym information */
	keysym->scancode = scancode;
	keysym->sym = keymap[scancode];
	keysym->mod = KMOD_NONE;

	/* If UNICODE is on, get the UNICODE value for the key */
	keysym->unicode = 0;
	if ( SDL_TranslateUNICODE ) {
		int map;
		SDLMod modstate;

		modstate = SDL_GetModState();
		map = 0;
		if ( modstate & KMOD_SHIFT ) {
			map |= (1<<KG_SHIFT);
		}
		if ( modstate & KMOD_CTRL ) {
			map |= (1<<KG_CTRL);
		}
		if ( modstate & KMOD_ALT ) {
			map |= (1<<KG_ALT);
		}
		if ( modstate & KMOD_MODE ) {
			map |= (1<<KG_ALTGR);
		}
		if ( KTYP(vga_keymap[map][scancode]) == KT_LETTER ) {
			if ( modstate & KMOD_CAPS ) {
				map ^= (1<<KG_SHIFT);
			}
		}
		if ( KTYP(vga_keymap[map][scancode]) == KT_PAD ) {
			if ( modstate & KMOD_NUM ) {
				keysym->unicode=KVAL(vga_keymap[map][scancode]);
			}
		} else {
			keysym->unicode = KVAL(vga_keymap[map][scancode]);
		}
	}
	return(keysym);
}
#endif

/* O. Kuss, D. Finkler */
extern void
FB_WriteFrontpanel (
	_THIS,
	const void *	iData,
	size_t			iLength)
{
	// Keyboard should already be open here
	if ( keyboard_fd > 0 ) {
		int result  = write (keyboard_fd, iData, iLength);
		int retries = 0;
		// Device may be busy, so give it some tries
		while ((result < 0) && (retries++ < 10))
		{
			// fprintf (stderr, "Write to frontpanel returned: %s\n", strerror (errno));
			// fprintf seems to cause Pizarro to crash when starting at boot time ?!?
			// Hhmmm, not explanation, so I guess anybody will read this once again :-)
			SDL_Delay(50);
			result = write (keyboard_fd, iData, iLength);
		}
		
		if (result < 0)
			SDL_SetError("Write to frontpanel failed");
	}
	else
		SDL_SetError("Write to frontpanel failed, keyboard not open");
}
